
//Title:        JeoEditUser
//Version:      1.2.0
//Copyright:    Copyright (c) 1998 Mark Grundland & Godfried Toussaint
//              All rights reserved.
//Author:       Mark Grundland

import java.applet.*;
import java.awt.*;

public abstract class JeoEditUser extends Applet implements Runnable {
/*
  This class allows other applets to communicate with JeoEdit editors.
  The user creates and edits the geometric objects,
    points in JeoEditPoints or a polygon in JeoEditPolgon,
      in the editor.
  The editor then acts as a server,
    sending the geometric information,
      a point set or a list of polygon vertices,
    to all the client applets that reside on
      the same web page as the editor.
  In order to receive this information, a client applet must be
    a subclass of JeoEditUser class and implement the receive method.

  Some Important Technical Points:

  * For the editor to be able to communicate with the client applets,
      they must reside on the same web page (i.e. the same HTML file) as
        as the editor,
      the web browser should be aware of their existance when the editor
        first starts running
        (this is a possible problem in very long HTML files),
      and the web browser must implement the AppletContext.getApplets()
        method correctly.

  * It is possible for multiple client applets to receive information
        from a single editor.
      Likewise, it is possible for multiple editors send information to
        a client applet.

  * The JeoEditUser class, a subclass of Applet,
      must be the superclass of every client applet.
      It overrides action(Event,Object) and getParameter(String) methods
        of the Applet class
      in order to enable communication
        between the editor and the client applet.
      If the client applet's subclass of JeoEditUser
        needs to override the action(Event,Object) method
        then it should use the following code:
          public boolean action(Event act,Object data){
            if (super.action(act,data)) {return true;}
            < process action event >
            ...
          }
      If the client applet's subclass of JeoEditUser
        needs to override the getParameter(String) method
        then it should use the following code:
          public String getParameter(String p){
            if (p.equals(JeoEditUser.usesJeoEdit)) {
              return super.getParameter(p);
            }
            < process parameter request >
            ...
          }
      Also the client applet's subclass of JeoEditUser
        should never override the run() method.


  * The editor will send the geometric information to a client applet,
      when the user invokes the Process Command
        (left click on the "GO" button in the editor),
      or, if the Auto-Process Option is turned on
        (right click on the "GO" button in the editor),
      whenever the user makes a change to the current geometric object.
      If the Auto-Verify Option is turned on
        (right click on the "Verify" button in the editor),
      the information will only be sent if it is deemed correct
        (a point set should contain no collinear points,
        while a polygon should be simple
          with no consecutive collinear vertices).

  * The client applet receives and processes geometric information
        in its own thread called requestThread,
      which is separate from any editor thread
        or the client's AWT event processing thread.
      The geometric information is not placed in a queue,
      hence if a client applet hasn't finished processing
        the last data which the editor sent,
      then the editor may forego sending new geometric data
        until the old data has been taken care of.
      If the client crashes due to an unhandled exception
        while processing the geometric information,
      its user interface and the editor should continue to work.
      In fact, the editor will try to restart the requestThread
        before sending more information.
      However notice that processing in the client's requestThread 
        can interfere with processing in the client's AWT event thread
        (the thread which handles the user's actions
          such as button clicks or key presses),
      so the client applet must be designed with these issues in mind.

  * Both JeoEditPoints and JeoEditPolgon describe their geometric data
      as x[] and y[] arrays of doubles,
      with each coordinate 0.0 <= x <= 1.0 and 0.0 <= y <= 1.0,
      and the origin is the top left corner
        (the usual Graphics origin in Java).
      These coordinates are not affected
        by the size of the editor's window;
        the geometric object is scaled
        to fit in the space provided for it.
      Hence editors usually display the geometric object
        with a rectangular aspect ratio,
        so the user edits the geometric object
        while it typically appears slightly horizontally stretched.
      Only JeoEditPoints ensures that no two lie on top of one another.
      With JeoEditPolgon, when the Auto-Verify Option is turned on
        and so the polygon is checked for simplicity before being sent,
      it is safe to assume that the data points denote the polygon
        vertices in counterclockwise order
      (if the polygon isn't simple you can only assume that the vertices
        are in consecutive order).

  * When the editor sends the geometric information
        to the client applet, its receive method will be invoked.
      In addition to the x[] and y[] coordinates of the geometric data,
        the recieve method is provided with
      the type of the data being sent (point set or polygon vertices),
      the number of the editor's workspace
        where the geometric object resides,
      the unique name of the editor sending the data,
      and the time when the data was sent.
      Hence the client applet can process the geometric data
        differently depending on the editor or workspace it comes from
        (e.g. the applet wishes to calculate the intersection of
          two polygons, so it stipulates that one polygon is
          created by the user in JeoEditPolygon's workspace 1
          and the other in workspace 2.)
*/

  public final static String usesJeoEdit="UsesJeoEdit";
    // the name of the applet parameter used indicate
    // that the applet wants to receive geometric data
    // from JeoEditPoints and JeoEditPolygon editors

  public final static int polygonDataID=400;
    // indicates that the geometric data describes a polygon
  public final static int pointsDataID=800;
    // indicates that the geometric data describes a polygon

  public Thread requestThread;
    // the thread which will respond to a request made by the editor
    // to process the current geometric object
    // by executing the receive method

  private final static int requestActionID=-12345;
  private final static String versionJeoEdit="100";
  private long lastRequestId=0^0;
  private Event requestEvent=
                new Event(null, 0, requestActionID, 0, 0, 0,0, null);
  private String requestSource="";
  private Object requestLock=new Object();
    // note: a request, in this context,
    // refers to the editor sending geometric data
    // for processing by the applet


  public JeoEditUser() {
  // initialize the handling of requests

    requestThread=new Thread(this,"JeoEdit Client");
    requestThread.start();
  }

  public String getParameter(String param){
  // signify that the applet should recieve requests from the editor

    if (param.equals(usesJeoEdit)) {return versionJeoEdit;}
    return super.getParameter(param);
  }

  public boolean action(Event act,Object data){
  // examine action events
  // to see if they transmit a request from the editor;
  // if so pass the request on to the request thread for processing;
  // note that this method may be called in the editor's thread

    if (act.id!=requestActionID) {return super.action(act,data);}
    if (!requestThread.isAlive()) {
      requestLock=new Object();
      requestThread.start();
    }
    synchronized (requestLock){
      requestEvent=act;
      requestSource=(String) data;
      requestLock.notify();
    }
    return true;
  }

  public void run(){
  // the main loop of the requestTread:
  // wait for a request to come from the editor
  // then translate the request
  // from an action event to an invocation of the receive method

    double[] x;
    double[] y;
    long timestamp;
    int workspace;
    int datatype;
    String editor;
    int reserved1;
    int reserved2;

    while (true){
      synchronized (requestLock){
        while (lastRequestId==(requestEvent.when^requestEvent.key)){
          try {requestLock.wait();} catch (InterruptedException exc) {;}
        }
        x=(double[]) requestEvent.target;
        y=(double[]) requestEvent.arg;
        timestamp=requestEvent.when;
        workspace=requestEvent.modifiers;
        datatype=requestEvent.key;
        editor=requestSource;
        reserved1=requestEvent.x;
        reserved2=requestEvent.y;
        lastRequestId=requestEvent.when^requestEvent.key;
      }
      receive(x,y,timestamp,workspace,datatype,editor,reserved1,reserved2);
    }
  }

  public abstract void receive (double x[],
                                double y[],
                                long timestamp,
                                int workspace,
                                int datatype,
                                String editor,
                                int reserved1,
                                int reserved2);
  // implement the recieve method to process the geometric data
  //  double x[]: 0.0 <= x[i] <= 1.0
  //    the x coordinates of the geometric data
  //  double y[]: 0.0 <= y[i] <= 1.0
  //    the y coordinates of the geometric data
  //  long timestamp: in milliseconds
  //    the time when the geometric data was sent
  //  int workspace: 1 <= workspace <= 5
  //    the editor's workspace where the geometric object resides
  //  int datatype: JeoEditUser constants {polygonDataID, pointsDataID}
  //    is polygonDataID if the data specifies the vertices of a polygon
  //    is pointDataID if the data specifies a point set
  //  String editor:
  //    unique name of the editor sending the data
  //  int reserved1:
  //    reserved for future use
  //  int reserved2:
  //    reserved for future use
}

